from inspect import getargvalues, signature, _empty
from os import getcwd
from re import compile
from sys import settrace
from time import perf_counter


class Tracer:
    def __init__(self,
        label       = 'tracing',
        truncate    = '_cwd_',
        show_types  = False,
        check_types = False,
        show_times  = False,
    ):
        self.label = label
        self.set_truncate(truncate)
        self.level = 0
        self.traces = list()
        self.show_types = show_types
        self.check_types = check_types
        self.ignored_paths = set()
        self.ignored_funcs = set()
        self.ignore_func('Tracer.__exit__')
        self.timers = dict()
        self.show_times = show_times

    def set_truncate(self, path):
        if path == '_cwd_':
           path = getcwd()
        self.truncate = len(path)+1

    def ignore_path(self, path):
        return self.ignored_paths.add(compile(path).search)

    def ignored_path(self, path):
        return any(ignore(path) for ignore in self.ignored_paths)

    def ignore_func(self, func):
        return self.ignored_funcs.add(compile(func).search)

    def ignored_func(self, func):
        return any(ignore(func) for ignore in self.ignored_funcs)

    def start_timer(self):
        if self.show_times:
            self.timers[self.level] = perf_counter()

    def stop_timer(self):
        if self.show_times:
            self.timers[self.level] = perf_counter() - self.timers[self.level]

    @property
    def elapsed(self):
        if self.show_times:
            elapsed = self.timers[self.level]
            if elapsed < 10:
                elapsed = int(elapsed * 1000)
                return f' [{elapsed} ms]' if elapsed > 0 else ''
            else:
                return f' [{round(elapsed)} s]'
        else:
            return ''

    def log(self, msg):
        tab = '    ' * self.level
        msg = f'{tab}{msg}'
        self.traces.append(msg)
        print(msg)

    def trace(self, file, line, func, args):
        if isinstance(args, tuple) and args[0] in ('>>>', '!!!'):
            if args[0] == '>>>':
                val = args[1]
                if self.show_types:
                    args = f' -> {type(val).__name__}: {val}{self.elapsed}'
                else:
                    args = f' -> {val}{self.elapsed}'
            elif args[0] == '!!!':
                args = f' ! {args[1].__name__} ! {args[2]}'
        else:
            if self.show_types:
                args = f"({', '.join(f'{k}:{type(v).__name__} = {v}' for k, v in args.items())})"
            else:
                args = f"({', '.join(f'{k}={v}' for k, v in args.items())})"
        self.log(f'{file}:{line} {func}{args}')


    def ann_err(self, name, expected, actual):
        if name == '_return_':
            msg = f'! "should return "{expected.__name__}" not "{actual.__name__}" !'
        else:
            msg = f'! arg "{name}" should be "{expected.__name__}" not "{actual.__name__}" !'
        self.log(msg)

    def get_func_args(self, frame):
        arginfo = getargvalues(frame)
        return {
            k:arginfo.locals[k]
            for k in arginfo.args
            if k != 'self'
        }

    def get_sig(self, frame):
        fct = frame.f_code.co_name
        slf = frame.f_locals.get('self')
        if slf:
            func = getattr(slf, fct)
        else:
            func = frame.f_globals[fct]
        return signature(func)

    def check_sig(self, frame, args):
        if self.check_types:
            sig = self.get_sig(frame)
            if isinstance(args, tuple) and args[0] == '>>>':
                a = sig.return_annotation
                if a is _empty: return
                if not isinstance(args[1], a):
                    self.ann_err('_return_', a, type(args[1]))
            else:
                for n, p in sig.parameters.items():
                    a = p.annotation
                    if a is _empty: continue
                    if isinstance(args, dict) and not isinstance(args[n], a):
                        self.ann_err(n, a, type(args[n]))

    def get_func_name(self, frame):
        fct = frame.f_code.co_name
        slf = frame.f_locals.get('self')
        func = getattr(slf, fct) if slf else frame.f_globals[fct]
        return func.__qualname__

    def get_trace(self, frame, arg):
        path = frame.f_code.co_filename[self.truncate:-3]
        if self.ignored_path(path): return
        func = self.get_func_name(frame)
        if self.ignored_func(func): return
        self.check_sig(frame, arg)
        line = frame.f_lineno
        self.trace(path, line, func, arg)

    def trace_exception(self, frame, typ, val, tbk):
        self.get_trace(frame, ('!!!', typ, val))

    def trace_return(self, frame, val):
        self.stop_timer()
        self.get_trace(frame, ('>>>', val))

    def trace_call(self, frame, event, arg):
        args = self.get_func_args(frame)
        self.get_trace(frame, args)
        self.start_timer()
        return self.trace_event

    def trace_event(self, frame, event, arg):
        if event == 'call':
            self.level+=1
            return self.trace_call

        if event == 'return':
            self.trace_return(frame, arg)
            self.level-=1

        if event == 'exception':
            self.trace_exception(frame, *arg)

    def dump(self):
        with open('trace.log', 'w') as out:
            out.write(f'{self.label} >>>\n')
            out.write('\n'.join(self.traces))
            out.write(f'\n{self.label} <<<')

    def __enter__(self):
        print(self.label, '>>>')
        settrace(self.trace_event)

    def __exit__(self, typ, val, tbk):
        print('<<<', self.label)
        settrace(None)
        self.dump()
        return True
